# -*- coding: utf-8 -*-
"""Document terminal here.

Copyright (C) 2021, Auto Trader UK
Created 13. Jan 2021 20:44

"""
from typing import List

import click

from olivertwist.ruleengine.result import Result


def get_colour(result: Result):
    colour = "green"
    if result.has_errors:
        colour = "red"
    if result.has_warnings:
        colour = "yellow"
    return colour


def report_to_terminal(results: List[Result]):
    for result in results:
        colour = get_colour(result)
        name = click.style(f"{result.rule.name}:", fg=colour)
        link = click.style(
            f"https://github.com/autotraderuk/oliver-twist/blob/main/RULES.md#{result.rule.id}",
            fg="blue",
        )
        click.echo(f"{name} [{link}]:")
        for node in result.failures:
            click.secho(f" - {node.id}", fg=colour)

        click.echo()



# -*- coding: utf-8 -*-
"""Models should all be enabled.

Copyright (C) 2020, Auto Trader UK
Created 15. Dec 2020 15:51

"""
from typing import List, Tuple, Optional

from olivertwist.manifest import Manifest, Node
from olivertwist.ruleengine.rule import rule
from olivertwist.rules.utils import partition


@rule(id="no-owner-on-physical-models", name="No owner on physical models")
def no_owner_on_physical_models(manifest: Manifest) -> Tuple[List[Node], List[Node]]:
    passes, failures = partition(
        lambda x: x.is_db_relation and __is_none_or_blank(x.owner), manifest.nodes()
    )
    return list(passes), list(failures)


def __is_none_or_blank(owner: Optional[str]):
    return owner is None or owner.strip(" ") == ""



# -*- coding: utf-8 -*-
"""Staging models should have a single source.

Copyright (C) 2020, Auto Trader UK
Created 22. Dec 2020 08:40

"""

from typing import List, Tuple

from olivertwist.manifest import Manifest, Node
from olivertwist.ruleengine.rule import rule
from olivertwist.rules.utils import partition


@rule(
    id="no-references-to-marts-from-staging", name="No references to marts from staging"
)
def no_references_to_marts_from_staging(
    manifest: Manifest,
) -> Tuple[List[Node], List[Node]]:
    def staging_depends_on_mart(node: Node):
        mart_refs = [
            p
            for p in manifest.graph.predecessors(node.id)
            if manifest.get_node(p).is_mart
        ]
        return node.is_staging and len(list(mart_refs)) > 0

    passes, failures = partition(staging_depends_on_mart, manifest.nodes())
    return list(passes), list(failures)



# -*- coding: utf-8 -*-
"""Auto-magic rule discovery.

Copyright (C) 2021, Auto Trader UK
Created 08. Jan 2021 18:01

"""
import pkgutil
from importlib import import_module

from olivertwist.ruleengine.rule import Rule


def rules_in_package(package):
    rules = []
    for info in pkgutil.walk_packages(
        path=package.__path__, prefix=package.__name__ + "."
    ):
        if "__" in info.name:
            continue

        mod = import_module(info.name)
        for obj in mod.__dict__.values():
            if isinstance(obj, Rule):
                rules.append(obj)

    return rules



# -*- coding: utf-8 -*-
"""Interactive rule configuration.

Copyright (C) 2021, Auto Trader UK
Created 08. Jan 2021 18:24

"""
from operator import attrgetter

from PyInquirer import prompt

import olivertwist
from olivertwist.config.model import Config, RuleConfig
from olivertwist.ruleengine.discovery import rules_in_package


class Configurator:
    @classmethod
    def update(cls, config: Config) -> Config:
        all_rules = rules_in_package(olivertwist)
        disabled = config.get_disabled_rule_ids()
        choices = [
            {
                "name": rule.name,
                "checked": rule.id not in disabled,
            }
            for rule in sorted(all_rules, key=attrgetter("name"))
        ]
        questions = [
            {
                "type": "checkbox",
                "name": "rules",
                "message": "Select rules to enable",
                "choices": choices,
                "validate": cls.__validate,
            }
        ]
        answers = prompt(questions)
        return Config(
            [
                RuleConfig(rule.id, False)
                for rule in all_rules
                if rule.name not in answers["rules"]
            ]
        )

    @classmethod
    def __validate(cls, answer):
        if len(answer) < 1:
            return "You must enable at least one rule."

        return True



# -*- coding: utf-8 -*-
"""Definition of the Rule class.

Copyright (C) 2020, Auto Trader UK
Created 15. Dec 2020 14:31

"""
from typing import Callable, List, Tuple

from olivertwist.config.model import Severity
from olivertwist.manifest import Manifest, Node


class Rule:
    def __init__(
        self,
        id: str,
        name: str,
        func: Callable[[Manifest], Tuple[List[Node], List[Node]]],
    ):
        self.id = id
        self.name = name
        self.func = func
        self._severity = Severity.ERROR

    def __call__(self, *args, **kwargs):
        return self.apply(*args, **kwargs)

    def apply(self, manifest) -> Tuple[List[Node], List[Node]]:
        return self.func(manifest)

    @property
    def severity(self):
        return self._severity

    @severity.setter
    def severity(self, severity: Severity):
        if not isinstance(severity, Severity):
            raise TypeError("severity must be an instance of Severity enum.")
        self._severity = severity


def rule(id, name):
    def class_constructor(func):
        nonlocal id, name
        if not name:
            name = func.__name__.replace("_", " ")

        return Rule(id, name, func)

    return class_constructor



# -*- coding: utf-8 -*-
"""Main entry point for olivertwist.

Copyright (C) 2020, Auto Trader UK
Created 15. Dec 2020 15:32

"""
import json
import logging
import os
import webbrowser
from pathlib import Path
from typing import List

import click

from olivertwist.config.configurator import Configurator
from olivertwist.config.io import ConfigIO
from olivertwist.manifest import Manifest
from olivertwist.metricengine.engine import MetricEngine
from olivertwist.reporter.adapter import to_html_report
from olivertwist.reporter.model import MyEncoder
from olivertwist.reporter.reporter import output_json, render_html_report
from olivertwist.reporter.terminal import report_to_terminal
from olivertwist.ruleengine.engine import RuleEngine
from olivertwist.ruleengine.result import Result

logger = logging.getLogger("olivertwist")


@click.group()
def main():
    pass


@main.command()
@click.option(
    "--config",
    "config_path",
    type=click.Path(dir_okay=False),
    help="The path to the configuration file to create / edit",
)
def config(config_path):
    """Interactively create or edit configuration file"""
    config = ConfigIO.read(config_path)
    config = Configurator.update(config)
    ConfigIO.write(config, Path(config_path or ConfigIO.DEFAULT_CONFIG_FILE_PATH))


@main.command()
@click.argument("input", type=click.File("r"))
@click.option(
    "--config", type=click.Path(exists=True), help="The path to the configuration file"
)
@click.option("--html/--no-html", default=True, help="Do/Don't output report in HTML")
@click.option(
    "--browser/--no-browser",
    default=False,
    help="Do/Don't open HTML report in browser. Implies --html",
)
def check(input, config, html=True, browser=False):
    """Check dbt DAG against configured rules."""
    config = ConfigIO.read(config)
    manifest = Manifest(json.load(input))
    rule_engine = RuleEngine.with_configured_rules(config)
    results = rule_engine.run(manifest)
    report_to_terminal(results)
    metric_results = MetricEngine().run(manifest)
    report = to_html_report(results, metric_results)
    oliver_twist = json.loads(MyEncoder().encode(report))
    output_json(oliver_twist)
    if html or browser:
        logger.debug("Generating HTML report...")
        render_html_report(oliver_twist)
        if browser:
            webbrowser.open(f"file://{os.getcwd()}/target/index.html")

    exit_message(results)


def format_for_terminal(results: List[Result]):
    for result in results:
        colour = get_colour(result)
        name = click.style(f"{result.rule.name}:", fg=colour)
        link = click.style(
            f"http://olivertwi.st/rules/#{result.rule.id}",
            fg="blue",
        )
        click.echo(f"{name} [{link}]:")
        for node in result.failures:
            click.secho(f" - {node.id}", fg=colour)

        click.echo()


def get_colour(result: Result):
    if result.has_errors:
        return "red"
    elif result.has_warnings:
        return "yellow"
    return "green"


def exit_message(results: List[Result]):
    if any([result.has_errors for result in results]):
        click.get_current_context().exit("🔀 Twisted!")
    else:
        click.echo("🟢 Oliver (all of your) models look good!")


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-
"""Staging models should have a single source.

Copyright (C) 2020, Auto Trader UK
Created 22. Dec 2020 10:29

"""

from typing import List, Tuple

from olivertwist.manifest import Manifest, Node
from olivertwist.ruleengine.rule import rule
from olivertwist.rules.utils import partition


@rule(
    id="no-references-outside-of-its-own-staging-area",
    name="No references outside of its own staging area",
)
def no_references_outside_of_its_own_staging_area(
    manifest: Manifest,
) -> Tuple[List[Node], List[Node]]:
    def staging_depends_on_staging_in_another_area(node: Node):
        different_staging_area_refs = [
            p
            for p in manifest.graph.predecessors(node.id)
            if manifest.get_node(p).is_staging
            if not manifest.get_node(p).area == node.area
        ]
        return node.is_staging and len(list(different_staging_area_refs)) > 0

    passes, failures = partition(
        staging_depends_on_staging_in_another_area, manifest.nodes()
    )
    return list(passes), list(failures)



# -*- coding: utf-8 -*-
"""Document __init__.py here.

Copyright (C) 2020, Auto Trader UK
Created 23. Dec 2020 13:40

"""

from dataclasses import dataclass
from enum import Enum
from typing import Optional, List

from dataclasses_jsonschema import JsonSchemaMixin


class Severity(Enum):
    WARNING = "warning"
    ERROR = "error"


@dataclass
class RuleConfig(JsonSchemaMixin):
    id: str
    enabled: bool
    severity: Optional[Severity] = Severity.ERROR


@dataclass
class Config(JsonSchemaMixin):
    universal: List[RuleConfig]

    def get_disabled_rule_ids(self) -> List[str]:
        return [r.id for r in self.universal if r.enabled is False]

    def get_config_for_rule_id(self, rule_id: str) -> RuleConfig:
        for rule_config in self.universal:
            if rule_config.id == rule_id:
                return rule_config

        return None



# -*- coding: utf-8 -*-
"""Document __init__.py here.

Copyright (C) 2020, Auto Trader UK
Created 22. Dec 2020 19:36

"""
from collections import Counter
from pathlib import Path
from typing import Union

import yaml

try:
    from yaml import CSafeLoader as YamlLoader
except ImportError:
    from yaml import SafeLoader as YamlLoader

from dataclasses_jsonschema import ValidationError

from olivertwist.config.model import Config


class InvalidConfigError(Exception):
    """Thrown if an invalid configuration file is supplied."""


class DuplicateEntryError(InvalidConfigError):
    """Duplicate sections were present in the supplied config."""


class ConfigIO:
    DEFAULT_CONFIG_FILE_PATH = Path("./olivertwist.yml")
    CONFIG_FILE_VERSIONS = {"1.0": Config}
    LATEST_CONFIG_FILE_VERSION = list(sorted(CONFIG_FILE_VERSIONS))[-1]

    @classmethod
    def read(cls, path: Union[Path, str]) -> Config:
        if path is None:
            if cls.DEFAULT_CONFIG_FILE_PATH.exists():
                config = cls.__parse(cls.DEFAULT_CONFIG_FILE_PATH)
            else:
                return Config(universal=[])
        else:
            config = cls.__parse(path)

        return cls.__validate(config)

    @classmethod
    def write(cls, config: Config, path: Union[Path, str]):
        with open(path, "w") as handle:
            config_dict = config.to_dict()
            config_dict["version"] = cls.LATEST_CONFIG_FILE_VERSION
            yaml.dump(config_dict, handle)

    @classmethod
    def __parse(cls, config_file_path: Union[Path, str]) -> Config:
        try:
            with open(config_file_path, "rb") as handle:
                yaml_config_dict = yaml.load(
                    handle.read().decode("utf-8"), Loader=YamlLoader
                )
                # When we have a later version we can do something with the version...
                _ = yaml_config_dict.pop("version")
                return Config.from_dict(yaml_config_dict)
        except KeyError:
            raise InvalidConfigError("Version is missing.")
        except ValidationError as e:
            raise InvalidConfigError(e)

    @classmethod
    def __validate(cls, config: Config) -> Config:
        entry_counts = Counter(rule.id for rule in config.universal)
        duplicates = [id_ for id_, count in entry_counts.items() if count > 1]
        if duplicates:
            raise DuplicateEntryError(", ".join(duplicates))

        return config



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['olivertwist',
 'olivertwist.config',
 'olivertwist.metricengine',
 'olivertwist.reporter',
 'olivertwist.ruleengine',
 'olivertwist.rules']

package_data = \
{'': ['*'],
 'olivertwist.reporter': ['html/css/*',
                          'html/images/*',
                          'html/webfonts/*',
                          'templates/*']}

install_requires = \
['Jinja2==2.11.2',
 'PyInquirer>=1.0.3,<2.0.0',
 'PyYAML>=5.3.1,<6.0.0',
 'click>=7.1.2,<8.0.0',
 'colorama<0.4.4',
 'dataclasses-jsonschema>=2.13.0,<3.0.0',
 'flake8-bugbear>=20.11.1,<21.0.0',
 'networkx>=2.5,<3.0']

extras_require = \
{':python_version >= "3.6" and python_version < "3.7"': ['dataclasses>=0.8,<0.9']}

entry_points = \
{'console_scripts': ['olivertwist = olivertwist.main:main',
                     'ot = olivertwist.main:main']}

setup_kwargs = {
    'name': 'olivertwist',
    'version': '0.2',
    'description': 'DBT DAG Auditor',
    'long_description': "\n\n![Alt text](https://github.com/autotraderuk/oliver-twist/blob/main/docs/images/oliver_twist_logo.png)\n# oliver-twist\n\nDAG Auditor\n\n[![Build status badge](https://github.com/autotraderuk/oliver-twist/workflows/CI/badge.svg)](https://github.com/autotraderuk/oliver-twist/actions?query=workflow%3ACI)\n[![Code style: black](https://img.shields.io/badge/code%20style-black-000000.svg)](https://github.com/psf/black)\n[![PyPI](https://img.shields.io/pypi/v/olivertwist)](https://pypi.org/project/olivertwist/)\n[![PyPI - Python Version](https://img.shields.io/pypi/pyversions/olivertwist)](https://pypi.org/project/olivertwist/)\n[![GitHub - License](https://img.shields.io/github/license/autotraderuk/oliver-twist)](https://github.com/autotraderuk/oliver-twist/blob/main/LICENSE)\n\noliver-twist is a dag auditing tool that audits the [DBT](https://www.getdbt.com/) DAG and generates a summary report. There is a [docs site][1], including descriptions of all [currently implemented rules][2].\n\n![please sir, can I automate my DAG auditing](https://github.com/autotraderuk/oliver-twist/blob/main/docs/images/oliver_dag_meme.jpg)\n\n# Getting Started\n\nTo get started, install the package\n\n```shell\n$ pip install olivertwist\n```\n\nand then run it by passing it your dbt manifest JSON\n\n```shell\nolivertwist check manifest.json\n```\n\nThis will report any failures to the console, and also in HTML format in a directory called `target`. You can optionally auto-open the report in a browser with:\n\n```shell\nolivertwist check manifest.json --browser\n```\n\nFull options are available with:\n\n\n```shell\nolivertwist check --help\n```\n\n## Configuration\n\n[All rules][2] are enabled by default. To change this you need a configuration file called `olivertwist.yml` in the same directory you are running `olivertwist`. An example configuration is shown below:\n\n```yaml\nversion: '1.0'\nuniversal:\n  - id: no-rejoin-models\n    enabled: false\n  - id: no-disabled-models\n    enabled: true\n```\n\nThere is a command to help you generate the config automatically:\n\n```shell\nolivertwist config\n```\n\nThis will show all the available rules and allow you to toggle the ones that you want to enforce.\n\n## Local Development\n\n### Clone this repo and install the project:\n\n`poetry install`\n\n### Install pre-commit hooks for linting\n\nThis is optional, but highly recommended to avoid annoying linting failure in CI.\n\n`poetry run pre-commit install`\n\nTo run the pre-commit hooks locally:\n\n`poetry run pre-commit run`\n\n### To get the latest versions of the dependencies and to update the poetry.lock file run:\n\n`poetry update`\n\n### To run oliver-twist and generate the summary report run:\n\n`poetry run olivertwist example_manifest.json`\n\n### Working with diagrams\n \nTo update and regenerate the images that illustrate rule failures in the documentation follow the next steps:\n- update the diagrams using the [mermaid syntax](https://mermaid-js.github.io/mermaid/#/)\n- install [yarn](https://classic.yarnpkg.com/en/docs/install/)\n- `cd docs/diagrams`\n- `./generate.sh`\n- inspect the generated images in `./docs/diagrams/output/`\n- if you're happy with the results, run `./copy.sh` so that they are copied over to `./docs/images`\n- you can now reference those images. e.g. in `.docs/rules.md`\n\n### Creating a distribution\n\n```poetry build --format wheel```\n\n\n[1]: http://olivertwi.st/\n[2]: http://olivertwi.st/rules/\n",
    'author': 'Angelos Georgiadis',
    'author_email': 'angelos.georgiadis@autotrader.co.uk',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'http://olivertwi.st',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)



# -*- coding: utf-8 -*-
"""Document engine here.

Copyright (C) 2020, Auto Trader UK
Created 15. Dec 2020 14:45

"""
from typing import List

import olivertwist
from olivertwist.config.model import Config, Severity
from olivertwist.manifest import Manifest
from olivertwist.ruleengine.discovery import rules_in_package
from olivertwist.ruleengine.result import Result
from olivertwist.ruleengine.rule import Rule


class RuleEngine:
    def __init__(self, rules: List[Rule]):
        self.rules = rules

    def __iter__(self):
        return iter(self.rules)

    def __get_enabled_rules(all_rules: List[Rule], config: Config):
        enabled_rules = []
        for rule in all_rules:
            if rule.id not in config.get_disabled_rule_ids():
                rule_config = config.get_config_for_rule_id(rule.id)
                if rule_config:
                    rule.severity = rule_config.severity
                else:
                    rule.severity = Severity.ERROR
                enabled_rules.append(rule)
        return enabled_rules

    @classmethod
    def with_configured_rules(cls, config: Config) -> "RuleEngine":
        all_rules = rules_in_package(olivertwist)
        enabled_rules = RuleEngine.__get_enabled_rules(all_rules, config)
        return cls(enabled_rules)

    def run(self, manifest: Manifest) -> List[Result]:
        return [Result(rule, *rule.apply(manifest)) for rule in self.rules]
